/**
	Bernard Fugas E1-ISIII
	
	Przykad wykorzystania szablonw generate oraz generate_n

	W skrcie:
	Szablony generate i generate_n pozwalaj przypisa wartoci zwracane
	przez bezargumentow funkcj lub obiekt funkcyjny do tablicy
	lub kontenera udostpniajcego iterator.
*/

#include <ostream>
#include <algorithm>	// tu jest definicja generate i generate_n
#include <vector>
#include "boost/random.hpp" //generator liczb losowych - uyty w przykadzie

using namespace std;

/** GeneticAlgorithm to klasa realizujca przykadowy algorytm genetyczny.
	Nie wnikamy w jego implementacj.

	Istotne jest to, e populacja pocztkowa skada si nie tylko z osobnikw
	wygenerowanych losowo, ale rwnie ze znanych "w miar dobrych" rozwiza
	zwanych "ekspertami".
	Szablony generate oraz generate_n pozwalaj w wygodny sposb stworzy
	populacj pocztkow o zadanej procentowej iloci ekspertw.

	Przykady uycia w funkcjach createGeneration0() 
	oraz operator<<(ostream& os, const GeneticAlgorithm& ga)
*/
class GeneticAlgorithm {
public:
	/** Przykadowa klasa osobnika uywanego w algorytmie genetycznym */
	class Individual {
		unsigned char chromosome_; //w rzeczywistym algorytmie byaby to co najmniej tablica bajtw
	public:
		Individual() {}
		Individual(unsigned char chromosome) : chromosome_(chromosome) {}
		friend class GeneticAlgorithm;
		friend ostream& operator<<(ostream& os, const Individual& i);
	};

	/** Obiekt funkcyjny generujcy liczby pseudolosowe */
	class RandomIndividualGenerator {
		boost::mt19937 state_;
		boost::uniform_int<unsigned> rng_;
	public:
		RandomIndividualGenerator(unsigned long state) : state_(state), rng_(0,255) {}
		Individual operator() (void) { return Individual( static_cast<unsigned char>(rng_(state_)) ); }
	};

////Definicje typw
	typedef vector<Individual> Individuals;
	typedef vector<Individual>::iterator IndividualsIt;
	typedef vector<Individual>::const_iterator IndividualsCit;

////Metody publiczne
	/**
	@brief Konstruktor algorytmu. 
	@param[in] population_size Rozmiar populacji
	@param[in] expert_percent  Procent ekspertw w populacji pocztkowej
	*/
	GeneticAlgorithm(unsigned long seed, size_t population_size, size_t expert_percent) :
		rig_(seed),
		population_size_(population_size), 
		expert_percent_(expert_percent)
		{}

	/** @brief Uruchamia szkielet algorytmu. */
	void run(void) {
		bool finish_condition = false;
		createGeneration0(); //stwrz populacj pocztkow

		while( !finish_condition ) {
			//korzystaj z operatorw genetycznych
			//selekcja, krzyowanie, mutacja, ...

			//w w ktrej z iteracji...
			finish_condition = true;
		}
	}

private:
	/**
	@brief Zwraca osobnikw "ekspertw" - znane, do dobre rozwizania
	@return Nowy osobnik "ekspert"
	@note Eksperci tworzeni s zgodnie z jak zalenoci funkcyjn.
	*/
	static Individual getExpert(void) {
		static unsigned char i = 0x77;
		++i;
		return Individual(i);
	}

	/**
	@brief Tworzy populacj pocztkow
	@note Przykad uycia generate
	*/
	void createGeneration0(void) {
		individuals_ = Individuals(population_size_);
		size_t experts_num = population_size_*expert_percent_/100; //liczba ekspertw
		IndividualsIt expertsEnd = individuals_.begin() + experts_num; //iterator "koca ekspertw"

		/**************************************************************************************
		void generate(_FwdIt _First, _FwdIt _Last, _Fn0 _Func)

		Pozwala przypisa elementom od iteratora _First (wcznie z _First) 
		do _Last (bez _Last) kolejne wartoci funkcji _Func
		**************************************************************************************/

		//generuj ekspertw
		//zgodnie z zalenoci funkcyjn; korzystajc z funkcji
		generate(individuals_.begin(), expertsEnd, getExpert); 
		
		//generuj losowe osobniki
		//losowo; korzystajc z funktora
		generate(expertsEnd, individuals_.end(), rig_);
		
		//mona sobie wyobrazi jeszcze inne rodzaje tworzonych w ten sposb osobnikw...
	}

	Individuals individuals_; /**< wektor osobnikw w populacji */
	size_t population_size_;  /**< rozmiar populacji */
	size_t expert_percent_;   /**< procentowy udzia ekspertw w populacji pocztkowej */
	RandomIndividualGenerator rig_; /**< generator losowych osobnikw (obiekt funkcyjny) */
	friend ostream& operator<<(ostream& os, const GeneticAlgorithm& ga);
	friend class Individual;
}; //class GeneticAlgorithm

/**
@brief Zaimplementowany w celu atwego wywietlenia rezultatu
*/
ostream& operator<<(ostream& os, const GeneticAlgorithm::Individual& individual) {
	os.setf(ios::hex, ios::basefield);
	os.width(2);
	os.fill('0');
	return os << static_cast<unsigned>(individual.chromosome_);
}

/**
@brief Wywietla biec populacj. 
Dla celw dydaktycznych rozpoczyna od wywietlenia sekwencji losowych osobnikw.
@note Przykad uycia generate_n
*/
ostream& operator<<(ostream& os, const GeneticAlgorithm& ga) {
	/**************************************************************************************
	generate_n(_OutIt _Dest, _Diff _Count, _Fn0 _Func)

	Pozwala przypisa n elementom poczwszy od _First 
	kolejne wartoci zwracane przez _Func

	Uwaga: tego nie da si wykona za pomoc generate
	**************************************************************************************/
	os << "Losowe liczby:" << endl;
	generate_n(ostream_iterator<GeneticAlgorithm::Individual>(os, " "), 10, ga.rig_);
	os << endl << "Populacja:" << endl;

	for(GeneticAlgorithm::IndividualsCit it=ga.individuals_.begin(); it != ga.individuals_.end(); ++it) 
		os << *it << endl;
	return os;
}

// ================== main ================ 
int main(/*int argc, char** argv*/) {
	size_t population_size = 20; /** < cakowity rozmiar populacji */
	size_t expert_percent = 10; /** < procentowa ilo ekspertw w populacji pocztkowej */
	unsigned long seed = 0;

	GeneticAlgorithm ga(seed, population_size, expert_percent); //stwrz algorytm
	ga.run();											  //i go uruchom
	//...
	cout << ga;	//wywietl populacj
}

